"""\
Purpose: Functions for parsing CSTA messages and creating CstaMessage objects
Initial Version: Costas Skarakis 11/11/2018
"""
import re
import xml.etree.ElementTree as ET
from csta.CstaMessage import CstaMessage

ENCODING = "utf8"


def buildMessageFromFile(filename, parameters, eventid):
    with open(filename, "r") as file:
        return buildMessage(file.read(), parameters, eventid)


def buildMessage(message, parameters={}, eventid=0):
    encoding_search = re.search("encoding=[\'\"](\S*)[\'\"].* ?\?\>", message)
    if not encoding_search:
        print("Warning: No encoding defined in message:", message)
        xml_encoding = ENCODING
    else:
        xml_encoding = encoding_search.group(1)
    tString = message.strip().format(**parameters)
    # bString=bytes(tString.replace("\n","\r\n")+2*"\r\n",encoding=xml_encoding)
    bString = bytes("%04d" % eventid + tString, encoding=xml_encoding)
    bHeader = bytes.fromhex("%08X" % (len(bString) + 4))
    cstaMessage = parseBytes(bHeader + bString)
    return cstaMessage


def parseBytes(bString):
    try:
        xml_encoding = re.search(b"encoding=[\'\"](\S*)[\'\"].* ?\?\>", bString).group(1)
        encoding = xml_encoding.decode(encoding=ENCODING)
    except AttributeError:
        encoding = ENCODING
    header = bString[:8]
    try:
        body = bString[8:].strip().decode(encoding)
    except:
        print("Cannot decode CSTA message with ", encoding)
        print(bString)
        raise
    try:
        root = ET.fromstring(body)
    except:
        print("Cannot parse CSTA message string:")
        print(header)
        print(body)
        raise
    tree = ET.ElementTree(root)
    # ns=re.search("^{(.*)}",root.tag)
    XMLParser = ET.XMLPullParser(events=['start-ns'])
    XMLParser.feed(body)
    ns = [e[1] for e in XMLParser.read_events()]
    if ns:
        # namespace=ns.group(1)
        cstamessage = CstaMessage(header, tree, body, encoding=encoding, ns=ns)
    else:
        # print("Warning: No namespace defined in message",root.tag)
        cstamessage = CstaMessage(header, tree, body, encoding=encoding)
    return cstamessage


if __name__ == "__main__":
    #    import os
    #    os.chdir(r'.\CstaPool')
    #    s=buildMessageFromFile("SystemRegister.xml",{"deviceID":10101001100,"reason":"because"},9999)
    a = []
    b = []
    a.append(b'\x00\x00\x03\xeb9999<?xml version="1.0" encoding="UTF-8"?><ServiceInitiatedEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1169730059</monitorCrossRefID><initiatedConnection><callID>FF000100000000002D65C85C1C750000</callID><deviceID>302118840100</deviceID></initiatedConnection><initiatingDevice><deviceIdentifier>N&lt;302118840100&gt;;noa=nk</deviceIdentifier></initiatingDevice><localConnectionInfo>initiated</localConnectionInfo><cause>makeCall</cause><servicesPermitted><callControlServices><clearConnection>true</clearConnection></callControlServices><callAssociatedServices/><mediaAttachementServices/><routeingServices/><voiceServices/></servicesPermitted><mediaCallCharacteristics><mediaClass><voice>true</voice><image>false</image><im>false</im></mediaClass></mediaCallCharacteristics><extensions><privateData><private xmlns:scx="http://www.siemens.com/schema/csta"><scx:extendedServicesPermitted/></private></privateData></extensions></ServiceInitiatedEvent>')
    a.append(b'\x00\x00\x04C9999<?xml version="1.0" encoding="UTF-8"?><OriginatedEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1169730059</monitorCrossRefID><originatedConnection><callID>FF000100000000002D65C85C1C750000</callID><deviceID>302118840100</deviceID></originatedConnection><callingDevice><deviceIdentifier>302118840100</deviceIdentifier></callingDevice><calledDevice><deviceIdentifier>N&lt;302118840101&gt;;displayNumber=302118840101;noa=nk</deviceIdentifier></calledDevice><localConnectionInfo>connected</localConnectionInfo><cause>newCall</cause><servicesPermitted><callControlServices><clearConnection>true</clearConnection></callControlServices><callAssociatedServices/><mediaAttachementServices/><routeingServices/><voiceServices/></servicesPermitted><mediaCallCharacteristics><mediaClass><voice>true</voice><image>false</image><im>false</im></mediaClass></mediaCallCharacteristics><extensions><privateData><private xmlns:scx="http://www.siemens.com/schema/csta"><scx:extendedServicesPermitted/></private></privateData></extensions></OriginatedEvent>')
    a.append(b'\x00\x00\x04A0002<?xml version="1.0" encoding="UTF-8"?><MonitorStop xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1193240031</monitorCrossRefID><extensions><privateData><private xmlns:scx="http://www.siemens.com/schema/csta"><scx:staticOndDN>17133013</scx:staticOndDN></private></privateData></extensions></MonitorStop>')
    a.append(b'\x00\x00\xb4A0002<?xml version="1.0" encoding="UTF-8"?><StartApplicationSessionPosResponse xmlns:aps="http://www.ecma-international.org/standards/ecma-354/appl_session"><sessionID>009968951</sessionID><actualProtocolVersion>http://www.ecma-international.org/standards/ecma-323/csta/ed4</actualProtocolVersion><actualSessionDuration>600</actualSessionDuration><extensions><actualHBT>0</actualHBT></extensions></StartApplicationSessionPosResponse>')
    b.append('''<?xml version="1.0" encoding="utf-8"?>
<MonitorStart>
   <monitorObject>
      <deviceObject>{user}</deviceObject>
   </monitorObject>
   <monitorType>device</monitorType>
   <extensions>
      <privateData>
         <private xmlns:scx="http://www.siemens.com/schema/csta">
            <scx:staticOndDN>17133013</scx:staticOndDN>
         </private>
      </privateData>
   </extensions>
</MonitorStart>''')
    a.append(b'\x00\x00\x04A0002<?xml version="1.0" encoding="UTF-8"?><SystemRegisterResponse xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><sysStatRegisterID>27ss</sysStatRegisterID></SystemRegisterResponse>')
    a.append(b'\x00\x00\x04A0002<?xml version="1.0" encoding="UTF-8"?><MonitorStartResponse xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1136370025</monitorCrossRefID></MonitorStartResponse>')
    b.append('''<?xml version="1.0" encoding="utf-8"?>
<StartApplicationSession>
   <applicationInfo>
      <applicationID>pythontsm</applicationID>
   </applicationInfo>
   <requestedProtocolVersions>
      <protocolVersion>http://www.ecma-international.org/standards/ecma-323/csta/ed4</protocolVersion>
   </requestedProtocolVersions>
   <requestedSessionDuration>600</requestedSessionDuration>
</StartApplicationSession>''')
    b.append('''<?xml version="1.0" encoding="UTF-8"?>
<SystemStatus
  xmlns="http://www.ecma.ch/standards/ecma-323/csta/ed2"
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
  xsi:schemaLocation="
      http://www.ecma.ch/standards/ecma-323/csta/ed2  file://localhost/X:/ips_bln/long_csta/ecma/system-status.xsd
  ">
  <systemStatus>normal</systemStatus>
</SystemStatus>''')
    b.append('''<?xml version="1.0" encoding="UTF-8" standalone="yes"?><ns3:SystemStatus xmlns:ns2="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns3="http://www.ecma-international.org/standards/ecma-323/csta/ed4"/>''')
    b.append('''<?xml version="1.0" encoding="UTF-8" standalone="yes"?><ns3:MonitorStart xmlns:ns2="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns3="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><ns3:monitorObject><ns3:deviceObject typeOfNumber="dialingNumber">3021033</ns3:deviceObject></ns3:monitorObject></ns3:MonitorStart>''')
    b.append('<?xml version="1.0" encoding="utf-8"?><DeflectCall><callToBeDiverted><deviceID>SUB_C</deviceID><callID>FF000100000000006BB1234C31000000</callID></callToBeDiverted><newDestination>SUB_D</newDestination></DeflectCall>')
    for i in a:
        print(parseBytes(i))
    for j in b:
        m = buildMessage(j, {"user": "12313213"}, eventid=2222)
        if m["deviceID"] and m["callID"]:
            print(m["deviceID"], m["callID"])
            m["deviceID"] = "user.deviceID"
            m["callID"] = "user.callID"
            print(m["deviceID"], m["callID"])
        print(m)
    last = buildMessage(j, {"user": "12313213"}, eventid=2222)["deviceObject"]





"""\
Purpose: Visualize test scenarios
Initial Version: Costas Skarakis 12/20/2021  
"""

import tkinter.tix as gr
import threading
from sip.SipEndpoint import SipEndpoint

MAXROWS = 40


class LoadWindow:
    def __init__(self):
        self.root = gr.Tk()
        scroll_window = gr.ScrolledWindow(self.root, width=300, height=800, scrollbar=gr.AUTO)
        scroll_window.pack(expand=1, fill=gr.BOTH)
        self.window = scroll_window.window
        self.subs = {}

    def paint(self, number):
        if number not in self.subs:
            count = len(self.subs)
            sub_row = count % MAXROWS
            sub_col = count // MAXROWS
            frame = gr.Frame(self.window)
            frame.grid(row=sub_row, column=sub_col)
            frame.label = gr.Label(frame, text=number, width=13)
            frame.arrow = gr.Label(frame, width=3, bg="yellow")
            frame.message = gr.Label(frame, width=30)
            frame.label.pack(side=gr.LEFT)
            frame.arrow.pack(side=gr.LEFT)
            frame.message.pack(side=gr.LEFT)
            self.subs[number] = frame

    def arrow(self, number, text):
        self.subs[number].arrow["text"] = text

    def message(self, number, text):
        self.subs[number].message["text"] = text

    def start(self, func_on_thread, *args, **kwargs):
        func_thread = threading.Thread(target=func_on_thread, args=args, kwargs=kwargs).start()
        self.root.mainloop()
        print("Exit main loop")
        func_thread.join()


class SipEndpointView(SipEndpoint):
    def __init__(self, view, number):
        self.view = view
        super().__init__(number)

    def connect(self, *args, **kwargs):
        self.view.paint(self.number)
        super().connect(*args, **kwargs)

    def send(self, *args, **kwargs):
        self.view.arrow(self.number, ">..")
        message = super().send(*args, **kwargs)
        self.view.arrow(self.number, "-->")
        self.view.message(self.number, message.get_status_or_method())
        return message

    def reply(self, *args, **kwargs):
        self.view.arrow(self.number, ">..")
        message = super().reply(*args, **kwargs)
        self.view.arrow(self.number, "-->")
        self.view.message(self.number, message.get_status_or_method())
        return message

    def send_new(self, *args, **kwargs):
        self.view.arrow(self.number, ">..")
        message = super().send_new(*args, **kwargs)
        self.view.arrow(self.number, "-->")
        self.view.message(self.number, message.get_status_or_method())
        return message

    def wait_for_message(self, message_type, **kwargs):
        self.view.arrow(self.number, "..<")
        self.view.message(self.number, message_type)
        message = super().wait_for_message(message_type, **kwargs)
        self.view.arrow(self.number, "<--")
        self.view.message(self.number, message.get_status_or_method())
        return message

    def colour(self, colour):
        self.view.subs[self.number].arrow["bg"] = colour

    def update_text(self, text=""):
        self.view.message(self.number, text)

    def update_arrow(self, arrow=""):
        self.view.arrow(self.number, arrow)



from tshark_tools.lib import get_from_cap_file
from sip.SipParser import buildMessage

L = get_from_cap_file("only_sip.pcap", tshark_path=r"c:\Program Files\Wireshark")

calls = []

for message in L:
    m = buildMessage(message, {})

    if m.get_status_or_method() == "INVITE":
        calls.append(m.header["Call-ID"])
    elif m["Call-ID"] in calls and m.type == "Response":
        assert m.status in ("200 OK", "100 Trying", "180 Ringing"), \
            "Validation error: Received:\n{}".format(m.contents())



from setuptools import setup

setup(
    name='pyssimist',
    version='0.1.3',
    packages=['sip', 'csta', 'common', 'tshark_tools'],
    url='https://github.com/skarcos/pyssimist',
    package_data={"csta": ["CstaPool/*.xml"]},
    license='GPL-3.0',
    author='Costas Skarakis',
    author_email='skarcos@gmail.com',
    description='A Python Network Traffic Simulator... It\'s probably not going to work.'
)



import argparse
import pprint
from datetime import datetime
from os import chdir, listdir, path
import sys
from string import ascii_uppercase
sys.path.append("..")
sys.path.append(path.join("..", ".."))
from tshark_tools.lib import get_msg_list_from_file, msg_filter, make_sip_message_template, summarize_trace
from sip.SipParser import buildMessage


def make_sub(i):
    max_letters = len(ascii_uppercase)
    numbering = str(int(i / max_letters))
    if numbering == "0":
        numbering = ""
    suffix = ascii_uppercase[i % max_letters] + numbering
    return "SUB_" + suffix


def get_sub(addr, addr_dict):
    if addr in addr_dict:
        return addr_dict[addr]
    else:
        sub = make_sub(len(addr_dict) + 1)
        addr_dict[addr] = sub
        return sub


def describe(tracefile, filters, wireshark_filter=None, tshark_path=None):
    """
    Describe a network trace as a runnable pyssimist test

    :param tracefile: Input network file in pcapng or json format
    :param wireshark_filter: Unless input is json, apply this wireshark filter to the output
    :param tshark_path: If not in PATH envirnonment variable, the path of the wireshark binaries must be given here
    :param filters: A list of filters to remove noise from messages.

    :return: A pyssimist testcase
    """
    tc_data = "tc_data.py"
    test_py = "test.py"
    if tracefile.endswith("json"):
        i_format = "json"
    else:
        i_format = "pcapng"
    count = 0
    subs = {}
    test_lines = "from tc_data import tc_message\n"
    test_data = {}
    summary = summarize_trace(tracefile, input_format=i_format, tshark_filter=wireshark_filter)
    for transport in summary["sip"]:
        if isinstance(summary["sip"][transport], list):
            for time_epoch, fromaddr, toaddr, message, expand in summary["sip"][transport]:
                a = get_sub(fromaddr, subs)
                b = get_sub(toaddr, subs)
                if "Call-ID" not in filters:
                    filters["Call-ID"] = []
                key = message.get_status_or_method().split()[0] + "_" + str(count)
                if msg_filter(message, filters) is None:
                    if message.type == "Request":
                        filters["Call-ID"].append(message["Call-ID"])
                    continue
                msg_template = make_sip_message_template(message, purpose="testcase")
                if msg_template not in test_data.values():
                    test_data[key] = msg_template
                    count += 1
                test_lines += "{}.send({}, tc_message['{}'])\n".format(a, b, key)
                test_lines += "{}.wait_for_message('{}')\n".format(b, message.get_status_or_method())

    with open(tc_data, "w") as sip_trace:
        sip_trace.write("tc_message = \\\n")
        sip_trace.write(pprint.pformat(test_data, width=200))
    with open(test_py, "w") as sip_test:
        sip_test.write(test_lines)


def GetArgs():
    """
    Supports the command-line arguments listed below.
    """
    parser = argparse.ArgumentParser(description='Summarize network capture file with expanded')
    parser.add_argument('-d', '--directory', required=False, action='store',
                        help='A directory containing the files to analyze')
    parser.add_argument('-i', '--input-file', required=False, action='store',
                        help='A directory containing the files to analyze')
    parser.add_argument('-f', '--filter', required=False, action='store',
                        help='Wireshark filter to apply to output')
    parser.add_argument('-t', '--tshark-path', required=False, action='store',
                        help='Wireshark binaries path. If omitted, it should be included in PATH environment variable]')
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = GetArgs()
    if args.directory:
        chdir(args.directory)
        for file in listdir("."):
            if file.rsplit(".", 1)[-1] in ("pcapng", "pcap", "cap", "json"):
                describe(file, filters={}, tshark_path=args.tshark_path, wireshark_filter=args.filter)
    if args.input_file:
        describe(args.input_file, filters={}, tshark_path=args.tshark_path, wireshark_filter=args.filter)


import os
import sys
import webbrowser
import ast
from os import path

sys.path.append("..")
sys.path.append(path.join("..", ".."))
from tshark_tools.diff import analyze
from tshark_tools.lib import check_in_trace, group_continuous_network_traces
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog, messagebox
from pprint import pformat
import threading


WIDTH = 1060
HEIGHT = 700


class Application:
    """Silly GUI"""

    def __init__(self):
        root = tk.Tk()
        root.title("Trace analyzer")
        self.master = root
        self.tests = {}
        self.tests_count = 0
        self.RunThread = None
        self.height = HEIGHT
        self.width = WIDTH

        style = ttk.Style()
        style.configure("help_style.Label", font=("Consolas", 9))
        style.configure("hyperlink.Label", foreground="blue", font=("Time", 9, "underline"))

        help_frame = ttk.Frame(self.master)
        help_frame.pack(side=tk.TOP)
        help_text = ttk.Label(help_frame, style="help_style.Label", text=check_in_trace.__doc__)
        help_text.pack()

        other_frames = ttk.Frame(self.master)
        other_frames.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)

        self.current_filter_frame = ttk.LabelFrame(other_frames, text="Import filter", width=self.width/3.0)
        self.current_filter_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)
        filter_preview = tk.StringVar()
        #ttk.Label(self.current_filter_frame, textvariable=filter_preview, style="help_style.Label").pack()
        self.import_filters = tk.Text(self.current_filter_frame, width=50, height=10, font=("Consolas", 9))
        self.import_filters.pack()
        #ttk.Button(self.current_filter_frame, text="Load Current", command=lambda: filter_preview.set(pformat(self.tests))).pack()
        ttk.Button(self.current_filter_frame, text="Load Current", command=self.load_current).pack()
        self.progress_bar = ttk.Progressbar(self.current_filter_frame, mode="indeterminate")

        self.command_frame = ttk.Frame(other_frames, width=self.width/3.0)
        self.command_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self.wireshark = tk.StringVar(value="sip||mgcp")
        ttk.Label(self.command_frame, text="Trace file selected:").pack()
        self.file_selected = ttk.Label(self.command_frame, text="No file selected")
        self.file_selected.pack()
        ttk.Button(self.command_frame, text="Select File", command=self.file_dialog).pack()
        doc_link = ttk.Label(self.command_frame, style="hyperlink.Label", text="Wireshark Display filter:")
        doc_link.bind("<Button-1>", lambda e: webbrowser.open_new("https://wiki.wireshark.org/DisplayFilters"))
        doc_link.pack()
        ttk.Entry(self.command_frame, textvariable=self.wireshark).pack()

        self.hide_unmatched = tk.BooleanVar(value=False)
        ttk.Radiobutton(self.command_frame, text="Hide unmatched", value=True, variable=self.hide_unmatched).pack()
        ttk.Radiobutton(self.command_frame, text="Collapse unmatched", value=False, variable=self.hide_unmatched).pack()
        self.run_button = ttk.Button(self.command_frame, text="Run", command=self.analyze_in_thread)
        self.run_button.pack(side=tk.BOTTOM)
        self.clear_button = ttk.Button(self.command_frame, text="Clear", command=self.clear_filters)
        self.clear_button.pack(side=tk.BOTTOM)

        filter_buttons_frame = ttk.Frame(self.command_frame)
        filter_buttons_frame.pack()
        ttk.Button(filter_buttons_frame, text="Add Header Filter (OR)", command=self.add_header_filter).pack(side=tk.LEFT)
        ttk.Button(filter_buttons_frame, text="Add SDP Filter (OR)", command=self.add_sdp_filter).pack(side=tk.LEFT)
        ttk.Button(filter_buttons_frame, text="Add XML Filter (OR)", command=self.add_xml_filter).pack(side=tk.LEFT)

        self.filters_frame = ttk.Frame(other_frames, width=self.width/3.0)
        self.filters_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)

    def load_current(self):
        self.import_filters.delete(1.0, tk.END)
        self.import_filters.insert(1.0, pformat(self.tests))

    def file_dialog(self):
        self.tests["Filename"] = filedialog.askopenfilename(initialdir=os.getcwd(),
                                                            title="Select A File",
                                                            filetypes=(("wireshark files", "*.cap *.pcap *.pcapng"),
                                                                       ("json files", "*.json"),
                                                                       ("all files", "*.*")),
                                                            multiple=True)
        if isinstance(self.tests["Filename"], str) or len(self.tests["Filename"]) == 1:
            self.file_selected.configure(text=self.tests["Filename"])
        else:
            self.file_selected.configure(text="Mulitple Selection")

    def clear_filters(self):
        self.tests_count = 0
        if "Filename" in self.tests:
            self.tests = {"Filename": self.tests["Filename"]}
        else:
            self.tests = {}
        filters = self.filters_frame.children
        for child_frame in filters:
            filters[child_frame].pack_forget()

    def add_header_filter(self):
        filter_frame = ttk.LabelFrame(self.filters_frame, text="Header criteria")
        filter_frame.pack(side=tk.LEFT)
        self.tests_count += 1
        self.tests["Test_header_%d" % self.tests_count] = {}
        self.add_headers(filter_frame, self.tests_count)
        return filter_frame

    def add_sdp_filter(self):
        filter_frame = ttk.LabelFrame(self.filters_frame, text="SDP criteria")
        filter_frame.pack(side=tk.LEFT)
        self.tests_count += 1
        self.tests["Test_sdp_%d" % self.tests_count] = {}
        self.add_sdp(filter_frame, self.tests_count)
        return filter_frame

    def add_xml_filter(self):
        filter_frame = ttk.LabelFrame(self.filters_frame, text="XML criteria")
        filter_frame.pack(side=tk.LEFT)
        self.tests_count += 1
        self.tests["Test_xml_%d" % self.tests_count] = {}
        self.add_xml(filter_frame, self.tests_count)
        return filter_frame

    def add_headers(self, frame, test_id):
        entries = []

        def set_header_values(a, b, c):
            self.import_filters.delete(1.0, tk.END)
            self.tests["Test_header_%d" % test_id]["Headers"] = {}
            for header, header_text in entries:
                if header.get():
                    self.tests["Test_header_%d" % test_id]["Headers"][header.get()] = header_text.get()

        def add_header():
            header = tk.StringVar()
            header_text = tk.StringVar()
            header.trace("w", set_header_values)
            header_text.trace("w", set_header_values)
            entries.append((header, header_text))
            ttk.Label(frame, text="Header:").pack()
            ttk.Entry(frame, textvariable=header).pack()
            ttk.Label(frame, text="Contains:").pack()
            ttk.Entry(frame, textvariable=header_text).pack()

        def set_request_value(a, b, c):
            self.import_filters.delete(1.0, tk.END)
            self.tests["Test_header_%d" % test_id]["Message"] = request.get()

        request = tk.StringVar()
        request.trace("w", set_request_value)
        self.tests["Test_header_%d" % test_id]["Headers"] = {}
        ttk.Label(frame, text="Msg type(empty=any):").pack()
        ttk.Entry(frame, textvariable=request).pack()
        ttk.Button(frame, text="Add header filter (AND)", command=add_header).pack()
        add_header()

    def add_sdp(self, frame, test_id):
        entries = []

        def set_sdp_values(a, b, c):
            self.import_filters.delete(1.0, tk.END)
            self.tests["Test_sdp_%d" % test_id]["sdp"] = {}
            for header, header_text in entries:
                sdp_line = header.get()
                if sdp_line:
                    if len(sdp_line) == 1:
                        sdp_line = sdp_line + "_line"
                    self.tests["Test_sdp_%d" % test_id]["sdp"][sdp_line] = header_text.get()

        def add_sdp_():
            header = tk.StringVar()
            header_text = tk.StringVar()
            header.trace("w", set_sdp_values)
            header_text.trace("w", set_sdp_values)
            entries.append((header, header_text))
            ttk.Label(frame, text="Line(o,a,v...):").pack()
            ttk.Entry(frame, textvariable=header).pack()
            ttk.Label(frame, text="Contains:").pack()
            ttk.Entry(frame, textvariable=header_text).pack()

        def set_sdp_includes_value(a, b, c):
            self.import_filters.delete(1.0, tk.END)
            self.tests["Test_sdp_%d" % test_id]["Message"] = sdp_includes.get()

        sdp_includes = tk.StringVar()
        sdp_includes.trace("w", set_sdp_includes_value)
        self.tests["Test_sdp_%d" % test_id]["sdp"] = {}
        ttk.Label(frame, text="Msg type(empty=any):").pack()
        ttk.Entry(frame, textvariable=sdp_includes).pack()
        ttk.Button(frame, text="Add sdp filter (AND)", command=add_sdp_).pack()
        add_sdp_()

    def add_xml(self, frame, test_id):
        entries = []

        def set_xml_values(a, b, c):
            self.import_filters.delete(1.0, tk.END)
            self.tests["Test_xml_%d" % test_id]["xml"] = {}
            for tag, tag_text, attr, attr_text in entries:
                xml_tag = tag.get()
                xml_attr = attr.get()
                xml_value = ""
                if xml_tag:
                    xml_value = xml_tag + " tag"
                self.tests["Test_xml_%d" % test_id]["xml"][xml_value] = tag_text.get()
                if xml_attr:
                    attr_value = "{} {} attr".format(xml_tag, xml_attr)
                    self.tests["Test_xml_%d" % test_id]["xml"][attr_value] = attr_text.get()

        def add_xml_():
            tag = tk.StringVar()
            attr = tk.StringVar()
            tag_text = tk.StringVar()
            attr_text = tk.StringVar()
            tag.trace("w", set_xml_values)
            tag_text.trace("w", set_xml_values)
            attr.trace("w", set_xml_values)
            attr_text.trace("w", set_xml_values)
            entries.append((tag, tag_text, attr, attr_text))
            ttk.Label(frame, text="XML label").pack()
            ttk.Entry(frame, textvariable=tag).pack()
            ttk.Label(frame, text="Contains in text:").pack()
            ttk.Entry(frame, textvariable=tag_text).pack()
            ttk.Label(frame, text="Or its attribute:").pack()
            ttk.Entry(frame, textvariable=attr).pack()
            ttk.Label(frame, text="Contains in value:").pack()
            ttk.Entry(frame, textvariable=attr_text).pack()

        def set_xml_includes_value(a, b, c):
            self.import_filters.delete(1.0, tk.END)
            self.tests["Test_xml_%d" % test_id]["Message"] = xml_includes.get()

        xml_includes = tk.StringVar()
        xml_includes.trace("w", set_xml_includes_value)
        self.tests["Test_xml_%d" % test_id]["xml"] = {}
        ttk.Label(frame, text="Msg type(empty=any):").pack()
        ttk.Entry(frame, textvariable=xml_includes).pack()
        ttk.Button(frame, text="Add xml filter (AND)", command=add_xml_).pack()
        add_xml_()

    def start(self):
        self.master.mainloop()

    def analyze_in_thread(self):
        self.run_button["state"] = tk.DISABLED
        self.RunThread = threading.Thread(target=self.analyze)
        self.RunThread.start()
        self.progress_bar.pack(fill=tk.X, expand=True)
        self.progress_bar.start(interval=10)
        self.master.after(1000, self.check_done)

    def check_done(self):
        if self.RunThread.is_alive():
            self.master.after(1000, self.check_done)
        else:
            self.RunThread.join()
            self.progress_bar.stop()
            self.progress_bar.pack_forget()
            self.run_button["state"] = tk.NORMAL
            self.master.update()

    def analyze(self):
        tests = self.tests
        imported_tests = self.import_filters.get(1.0, tk.END).strip()
        if imported_tests:
            tests = ast.literal_eval(imported_tests)
        filters = [tests[test] for test in tests if test.startswith("Test")]
        if isinstance(self.tests["Filename"], str):
            file_generated = analyze(self.tests["Filename"], *filters, hide_unmatched=self.hide_unmatched.get(),
                                     wireshark_filter=self.wireshark.get())
            messagebox.showinfo(title="Trace analyzed", message="Output filename: %s" % file_generated)
        else:
            file_groups = group_continuous_network_traces(self.tests["Filename"])
            files_generated = ""
            for group in file_groups:
                files_generated += analyze(file_groups[group], *filters, hide_unmatched=self.hide_unmatched.get(),
                                           wireshark_filter=self.wireshark.get()) + "\n"
            messagebox.showinfo(title="Trace analyzed", message="Output filenames: %s" % files_generated)


if __name__ == "__main__":
    Application().start()



"""\
Purpose: Define reusable SIP Flows
Initial Version: Costas Skarakis 5/25/2021
"""
from time import sleep


def basic_call_monitored_events(A, B, duration=2):
    """ Basic CSTA Call flow between two CSTA Monitored Endpoints
    SIP flow must be handled elsewhere.
    SIP_A calls SIP_B. B answers. SIP_A hangs up after the duration.
    CSTA Users A, B will receive the corresponding CSTA events

    :A: The caller (CstaUser instance)
    :B: The callee (CstaUser instance)
    :duration: The duration of the call
    """
    Aside = ["ServiceInitiatedEvent", "OriginatedEvent", "DeliveredEvent", "EstablishedEvent",
             "ConnectionClearedEvent"]
    Bside = ["DeliveredEvent", "EstablishedEvent", "ConnectionClearedEvent"]

    A.wait_for_message("ServiceInitiatedEvent")
    A.wait_for_message("OriginatedEvent")
    A.wait_for_message("DeliveredEvent")
    A.wait_for_message("EstablishedEvent")

    B.wait_for_message("DeliveredEvent")
    B.wait_for_message("EstablishedEvent")

    sleep(duration)

    B.wait_for_message("ConnectionClearedEvent")

    A.wait_for_message("ConnectionClearedEvent")



"""\
Purpose: Network connection facilities
Initial Version: Costas Skarakis 11/11/2018
"""
import selectors
import socket
import ssl
import io
from threading import Lock

from common.tc_logging import debug
from common.util import wait_for_sip_data, NoData


class TCPClient(object):
    def __init__(self, ip, port):
        self.sip_buffer = []
        self.ip = ip
        self.port = port
        self.rip, self.rport = None, None
        if ":" in self.ip:
            # ipv6 case
            self.socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM, 0)
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.send_lock = Lock()
        self.wait_lock = Lock()
        self.sel = selectors.DefaultSelector()
        self.sel.register(self.socket, selectors.EVENT_READ, data=None)

    def connect(self, dest_ip, dest_port):
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((self.ip, self.port))
        self.port = self.socket.getsockname()[1]
        self.socket.settimeout(5.0)
        self.sockfile = self.socket.makefile(mode='rb')

        self.socket.connect((dest_ip, dest_port))
        self.rip = dest_ip
        self.rport = dest_port

    def send(self, data, encoding="utf8"):
        # self.socket.sendall(binascii.hexlify(bytes(data,"utf8")))
        with self.send_lock:
            if type(data) == type(b''):
                self.socket.sendall(data)
                debug("Sent from port {}:\n\n".format(self.port) + data.decode("utf8", "backslashreplace").replace("\r\n",
                                                                                                                   "\n"))
            else:
                self.socket.sendall(bytes(data, encoding))
                debug("Sent from port {}:\n\n".format(self.port) + data.replace("\r\n", "\n"))

    def waitForData(self, timeout=None, buffer=4096):
        debug("Waiting on port {}".format(self.port))
        bkp = self.socket.gettimeout()
        if timeout:
            self.socket.settimeout(timeout)
        try:
            data = self.socket.recv(buffer)
        finally:
            self.socket.settimeout(bkp)
        debug("Received on port {}:\n\n".format(self.port) + data.decode("utf8", "backslashreplace").replace("\r\n",
                                                                                                             "\n"))
        return data

    def wait_select(self, timeout):
        if not self.sel.select(timeout):
            raise socket.timeout("Timeout waiting for data in " + self.ip + ":" + str(self.port))

    def waitForSipData(self, timeout=None, client=None, bufsize=4096):
        if not client:
            client = self
        with client.wait_lock:
            debug("Waiting on port {}".format(client.port))
            bkp = client.socket.gettimeout()
            data = b""
            try:
                if client.sip_buffer:
                    data = client.sip_buffer.pop(0)
                else:
                    client.wait_select(timeout)
                    all_data = io.BytesIO(client.socket.recv(bufsize))
                    while True:
                        try:
                            # There may be more than one sip messages in the socket
                            # I think select will not trigger if we only read one of them
                            # and then come back for the second one
                            client.sip_buffer.append(wait_for_sip_data(all_data))
                        except NoData:
                            if client.sip_buffer:
                                data = client.sip_buffer.pop(0)
                            else:
                                data = all_data.getvalue()
                                raise ValueError
                            break
            except ValueError:
                debug('Value Error: '+data.decode())
                # debug(line.decode())
                raise
            except socket.timeout:
                debug('Data received before timeout: "{}"'.format(data.decode()))
                raise
            finally:
                client.socket.settimeout(bkp)
            debug("Received on port {}:\n\n".format(client.port) + data.decode("utf8").replace("\r\n", "\n"))
            return data

    def waitForCstaData(self, timeout=None):
        with self.wait_lock:
            bkp = self.socket.gettimeout()
            if timeout: self.socket.settimeout(timeout)
            try:
                # header = ""
                # while not header:
                header = self.socket.recv(4)
                if not header:
                    # disconnected socket
                    return None
                datalength = int(''.join(["%02X" % x for x in header]), base=16) - 4
                data = b''
                while len(data) < datalength:
                    data += self.socket.recv(datalength - len(data))
                debug(
                    "Received on port {} message of length {}:\n\n".format(self.port, datalength) + (header + data).decode(
                        "utf8",
                        "backslashreplace").replace("\r\n", "\n"))
            finally:
                self.socket.settimeout(bkp)
            return header + data


class UDPClient(TCPClient):
    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.rip, self.rport = None, None
        if ":" in self.ip:
            # ipv6 case
            self.socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM, 0)
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((self.ip, self.port))
        self.socket.settimeout(5.0)
        self.sockfile = self.socket.makefile(mode='rb')
        self.send_lock = Lock()
        self.wait_lock = Lock()


class TLSClient(TCPClient):
    def __init__(self, ip, port, certificate=None, subject_name="localhost"):
        self.ip = ip
        self.port = port
        self.rip, self.rport = None, None
        self.server_name = subject_name
        self.send_lock = Lock()
        self.wait_lock = Lock()

        # PROTOCOL_TLS_CLIENT requires valid cert chain and hostname
        if hasattr(ssl, "PROTOCOL_TLS_CLIENT"):
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        elif hasattr(ssl, "PROTOCOL_TLSv1_1"):
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_1)
        else:
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)

        if not certificate:
            self.context.check_hostname = False
            self.context.verify_mode = ssl.CERT_NONE
            ssl.create_default_context()
        else:
            # certificate example: 'path/to/my_certificate.pem'
            self.context.load_verify_locations(certificate)

    def connect(self, dest_ip, dest_port):
        if ":" in self.ip:
            # ipv6 case
            tcp_socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM, 0)
        else:
            tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket = self.context.wrap_socket(tcp_socket, server_hostname=self.server_name)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
#        self.socket.bind((self.ip, self.port))
        #self.port = self.socket.getsockname()[1]
        self.socket.settimeout(5.0)
        self.sockfile = self.socket.makefile(mode='rb')
        super().connect(dest_ip, dest_port)
